#include <hwloc.h>
#include "ReadTopo.h"


void navigate_topology(hwloc_topology_t topology,hwloc_obj_t obj,struct multicore_topo * topo_info){
	switch (obj->type) {
		case HWLOC_OBJ_CACHE:
/*			printf("cache size %lu", obj->attr->cache.memory_kB);*/
/*			printf("cache depth %u\n", obj->attr->cache.depth);*/

			if(obj->arity>1 && !topo_info->is_cs_set){
				topo_info->cs = obj->attr->cache.memory_kB;
				topo_info->is_cs_set = 1;
			}
			else if(obj->arity==1 && !topo_info->is_cd_set){
				topo_info->cd = obj->attr->cache.memory_kB;
				topo_info->is_cd_set = 1;
			}
		break;

		case HWLOC_OBJ_NODE:
/*			printf("Node mem %lu", obj->attr->node.memory_kB);*/
/*			printf("huge page free %lu\n", obj->attr->node.huge_page_free);*/
			topo_info->node_count++;
		break;

		case HWLOC_OBJ_CORE:
/*			printf("\t\tCore %u\n", obj->logical_index);*/
			topo_info->core_count++;
		break;

		case HWLOC_OBJ_PROC:
/*			printf("\tProc %u\n", obj->logical_index);*/
			topo_info->proc_count++;
		break;

		default:
		break;
	}

	if (obj->arity) {
		int x;
		for (x=0; x<obj->arity; x++)
			navigate_topology (topology, obj->children[x],topo_info);
  	}
}

int ReadTopology(struct multicore_topo * topo_info){
	int err;
	hwloc_topology_t topology;
  //	struct hwloc_topology_info info;
	unsigned long flags = 0;
	err = hwloc_topology_init (&topology);
	hwloc_topology_set_flags(topology, flags);

	err = hwloc_topology_load (topology);
	if (err)
		return 1;


//	hwloc_topology_get_info(topology, &info);

	topo_info->is_cd_set = 0;
	topo_info->is_cs_set = 0;
	topo_info->node_count = 0;
	topo_info->proc_count = 0;
	topo_info->core_count = 0;
	topo_info->cs = 0;
	topo_info->cd = 0;

	navigate_topology(topology,hwloc_get_system_obj(topology),topo_info);

	hwloc_topology_destroy (topology);

	return 0;
}
