/*
 *  tracecreator.cpp
 *  cache_sim
 *
 *  Created by mathias jacquelin on 1/23/09.
 *  Copyright 2009 LIP/ENS-Lyon. All rights reserved.
 *
 */
#include <mkl_cblas.h>
#include <mkl_blas.h>

#include <string.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/shm.h>
#include <sys/ipc.h>

#include <numa.h>
#include <math.h>
#include <getopt.h>
#include <time.h>
#include <sys/time.h>

#include "ReadTopo.h"

#ifdef GOTO
#include <common.h>
#define dgemm   BLASFUNC(dgemm)
#endif

#ifndef PAGESIZE
#define PAGESIZE    ( 4 << 10 )
#endif
#define HUGE_PAGESIZE   ( 2 << 20 )

#define CACHE_PROP_CS 1.0
#define CACHE_PROP_CD 0.7


#ifdef ROWMAJOR
#define A(I,J) a[((I))*n+(J)]
#define B(I,J) b[((I))*n+(J)]
#define C(I,J) c[((I))*n+(J)]
#define D(I,J) d[((I))*n+(J)]
#else
#define A(I,J) a[((J))*n+(I)]
#define B(I,J) b[((J))*n+(I)]
#define C(I,J) c[((J))*n+(I)]
#define D(I,J) d[((J))*n+(I)]
#endif


#ifndef max
#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

#ifndef min
#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif

enum algo_type {
    OUTER,
    MMRE_S,
    MMRE_SLRU,
    MMRE_D,
    MMRE_DLRU,
    MMRE_T,
    MMRE_TLRU,
    MMRE_TM,
    MMRE_TMLRU,
    TOLEDO,
    TOLEDO_LRU,
    TOLEDO_S,
    TOLEDO_SLRU,
    LIBRARY
};

#ifdef MYDGEMM
my_dgemm(const int* m,const int* n,const int* p, const double *alpha, const double *a,const int* lda,const double * b,const int* ldb,const double* beta,double * c,const int* ldc){
    int i,j,k;
    for(i=0;i<*m;++i){
        for(j=0;j<*n;++j){
        double tmp = c[i**ldc+j];
            for(k=0;k<*p;++k){
                    tmp += a[i**lda + k]* b[k**ldb + j];
                }
        c[i**ldc+j] = tmp;
        }
    }
};

#define dgemm(CblasATrans,CblasBTrans,m,n,p,alpha,a,lda,b,ldb,beta,c,ldc) my_dgemm(m,n,p,alpha,a,lda,b,ldb,beta,c,ldc)

#define cblas_dgemm(CblasXMajor,CblasATrans,CblasBTrans,m,n,p,alpha,a,lda,b,ldb,beta,c,ldc) my_dgemm(&m,&n,&p,&alpha,a,&lda,b,&ldb,&beta,c,&ldc)
#endif

int get_lambda(int cs){
    return max((int) floor(sqrt((double)cs -0.75) - 0.5),1);
}

int get_alpha_max(int cs){
    return max((int) floor(-1.0 + sqrt((double)cs + 1.0)),1);
}


int get_mu(int cd){
    return max((int) floor(sqrt((double)cd -0.75) - 0.5),1);

}

#define ADD_INV_MOD_P(x,p) (((int)(p) -( (int)(x) % (int)(p))) % (int)(p) )
void get_alpha_beta(int cs, double bws, double bwd, int p,int mu, int block,int  n,int * alpha, int * beta){
	*alpha = floor(sqrt((double)cs*(1.0+2.0*(double)p*bwd/bws - sqrt(1.0+8.0*(double)p*bwd/bws))/(2.0*(double)p*bwd/bws -2.0)));

	*alpha += ADD_INV_MOD_P(*alpha,(int)sqrt(p)*mu);
	*alpha -= (int)sqrt(p)*mu;

		while(*alpha>sqrt(p)*mu && (*alpha%((int)sqrt(p)*mu)!=0 || n%(*alpha*block)!=0) )
		{
			--*alpha;
		}


	*beta = ((double)cs - (double)pow(*alpha,2))/(4.0**alpha);
	while(*beta>1 && n%(*beta*block)!=0  )
		--*beta;
}


int read_arguments(int argc, char *argv[],int * blocksize, enum algo_type * algorithm_type, int * n) {
    int c;

    while (1) {
        static struct option long_options[] = {
            /* These options set a flag.   */
            { "help", no_argument, 0, 'h' },
            { "n", required_argument, 0, 'n' },
            { "block", required_argument, 0, 'b' },
            { "algo", required_argument, 0, 'a' },
            { 0, 0, 0, 0 }
        };
        /* getopt_long stores the option index here.   */
        int option_index = 0;
        c = getopt_long(argc, argv, "hb:l:", long_options, &option_index);


        /* Detect the end of the options.   */
        if (c == -1)
            break;


        switch (c) {
            case 'h':
                printf("Usage: %s [options] file\n", argv[0]);
                exit(0);
                break;
            case 'b':
                *blocksize = atoi(optarg);
                break;
            case 'n':
                *n = atoi(optarg);
                break;
            case 'a':
                {
                    if(strcmp(optarg,"o")==0)
                    {
                        *algorithm_type = OUTER;
                    }
                    else if(strcmp(optarg,"d")==0)
                    {
                        *algorithm_type = MMRE_D;
                    }
                    else if(strcmp(optarg,"dlru")==0)
                    {
                        *algorithm_type = MMRE_DLRU;
                    }
                    else if(strcmp(optarg,"s")==0)
                    {
                        *algorithm_type = MMRE_S;
                    }
                    else if(strcmp(optarg,"slru")==0)
                    {
                        *algorithm_type = MMRE_SLRU;
                    }
                    else if(strcmp(optarg,"t")==0)
                    {
                        *algorithm_type = MMRE_T;
                    }
                    else if(strcmp(optarg,"tlru")==0)
                    {
                        *algorithm_type = MMRE_TLRU;
                    }
                    else if(strcmp(optarg,"tm")==0)
                    {
                        *algorithm_type = MMRE_TM;
                    }
                    else if(strcmp(optarg,"tmlru")==0)
                    {
                        *algorithm_type = MMRE_TMLRU;
                    }
                    else if(strcmp(optarg,"tol")==0)
                    {
                        *algorithm_type = TOLEDO;
                    }   
                    else if(strcmp(optarg,"tollru")==0)
                    {
                        *algorithm_type = TOLEDO_LRU;
                    }
                    else if(strcmp(optarg,"tol_s")==0)
                    {
                        *algorithm_type = TOLEDO_S;
                    }
                    else if(strcmp(optarg,"tol_slru")==0)
                    {
                        *algorithm_type = TOLEDO_SLRU;
                    }
                    else if(strcmp(optarg,"lib")==0)
                    {
                        *algorithm_type = LIBRARY;
                    }

                }

                break;
            default:
                exit(0);
        }

    }
    return 0;
}



void matrix_multiplication(const double *ma,const double *mb, double *mc,int mn, int numproc,int blocksize)
{
    double * a;
    double * b;
    double * c;
    int n;
    int blk_size;

    int proc;
    int ii,jj,k,i,j;
    int mi,mj;
    int step;
    int maxij;
    int maxk;
    int sqrtproc;
    double alpha;

    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;

    alpha = 1.0;
    sqrtproc = floor(sqrt(numproc));    
    step = sqrtproc;
    maxij = ceil((float)n/((float)step*blocksize));
    maxk = ceil((float)n/((float)blocksize));

#ifdef ROWMAJOR
#pragma omp parallel firstprivate(alpha,a,b,c,n,blk_size,sqrtproc,step,maxij,maxk) private(ii,jj,k,mi,mj,proc,i,j)
    {
#pragma omp for schedule(static,1) 
        for(proc=1;proc<=numproc;proc++)
        {

            for (k = 0; k<maxk ; ++k) {
                for (ii = 0; ii < maxij; ++ii) {
                    i = ii*step;
                    mi = i + ((proc-1) % sqrtproc );
                    for (jj = 0; jj < maxij; ++jj) {
                        j = jj*step;
                        mj = j + ((proc-1) / sqrtproc );
                        dgemm("N","N",&blocksize,&blocksize,&blocksize,&alpha,&A(mi*blocksize,k*blocksize),&n,&B(k*blocksize,mj*blocksize),&n,&alpha,&C(mi*blocksize,mj*blocksize),&n);
#ifdef SYNC
                        #pragma omp barrier
#endif
                    }
                }
            }

        }
    }
#else
#pragma omp parallel firstprivate(alpha,a,b,c,n,blk_size,sqrtproc,step,maxij,maxk) private(ii,jj,k,mi,mj,proc,i,j)
    {
#pragma omp for schedule(static,1) 
        for(proc=1;proc<=numproc;proc++)
        {

            for (k = 0; k<maxk ; ++k) {
                for (jj = 0; jj < maxij; ++jj) {
                    j = jj*step;
                    mj = j + ((proc-1) / sqrtproc );
                    for (ii = 0; ii < maxij; ++ii) {
                        i = ii*step;
                        mi = i + ((proc-1) % sqrtproc );
                        dgemm("N","N",&blocksize,&blocksize,&blocksize,&alpha,&A(mi*blocksize,k*blocksize),&n,&B(k*blocksize,mj*blocksize),&n,&alpha,&C(mi*blocksize,mj*blocksize),&n);
                    }
                }
            }
        }
    }
#endif
}


void MMRE_shared(const double *ma,const double *mb, double *mc,int mn, int numproc, int cs, int blocksize)
{
    //need to find parameters lambda and mu
    const double * a;
    const double * b;
    double * c;
    int proc_count,blk_size,n;
    int lambda,maxij,maxk,inner_step,hdim;
    int i,j,bj,p,stepi,stepj,k,ia,my_j;
    int bbj;
    double alpha;

    /*copying parameters in private variables*/ 
    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;
    proc_count = numproc;

    lambda = get_lambda(cs);

    /*lambda must be a multiple of both core count and matrix order:*/
    int lmb = lambda;
    while(lmb>0 && (lmb % numproc != 0 || n%(lmb*blocksize) !=0) )
    {
        lmb--;
    }

    lambda = max(lmb,numproc);

    /*compute maxij and maxk (upper integer part)*/
    maxij = ceil((double)mn/((double)lambda*(double)blocksize));
    maxk = ceil((double)mn/((double)blocksize));

    inner_step = floor((float)lambda/(float)numproc);
    hdim = blocksize*inner_step;
    alpha = 1.0;

#ifdef ROWMAJOR
#pragma omp parallel for schedule(static,1) firstprivate(a,b,c,n,blk_size,proc_count,lambda,maxij,maxk,inner_step,alpha,hdim) private(my_j,i,j,bj,p,stepi,stepj,k,ia) private(bbj)
        for(p = 1; p<=proc_count ;++p)
        {
            my_j = (p-1) *inner_step;
            for(stepi = 0; stepi <maxij; ++stepi)
            {
                i = stepi*lambda;
                for(stepj = 0; stepj <maxij; ++stepj)
                {
                    j = stepj*lambda;           
                    bj=j+my_j;
                    for(k=0; k<maxk;++k)
                    {
                        for(ia = i; ia< i+lambda; ++ia)
                        {
                            
                            for(bbj=0;bbj<inner_step;++bbj){
                                dgemm("N","N",&blk_size,&blk_size,&blk_size,&alpha,&B(k*blk_size,(bj+bbj)*blk_size),&n,&A(ia*blk_size,k*blk_size),&n,&alpha,&C(ia*blk_size,(bj+bbj)*blk_size),&n);
                                #ifdef SYNC
                            #pragma omp barrier
#endif

                            }
                        }
                    }
                }
            }
        }
//    }
#else
#pragma omp parallel for schedule(static,1) firstprivate(a,b,c,n,blk_size,proc_count,lambda,maxij,maxk,inner_step,alpha,hdim) private(my_j,i,j,bj,p,stepi,stepj,k,ia) private(bbj)
        for(p = 1; p<=proc_count ;++p)
        {
            my_j = (p-1) *inner_step;
            for(stepj = 0; stepj <maxij; ++stepj)
            {
                j = stepj*lambda;           
                bj=j+my_j;
                for(stepi = 0; stepi <maxij; ++stepi)
                {
                    i = stepi*lambda;
                    for(k=0; k<maxk;++k)
                    {
                        for(ia = i; ia< i+lambda; ++ia)
                        {
                            
                            for(bbj=0;bbj<inner_step;++bbj){
                                dgemm("N","N",&blk_size,&blk_size,&blk_size,&alpha,&A(ia*blk_size,k*blk_size),&n,&B(k*blk_size,(bj+bbj)*blk_size),&n,&alpha,&C(ia*blk_size,(bj+bbj)*blk_size),&n);
                                #ifdef SYNC
                            #pragma omp barrier
#endif
                            }
                        }
                    }
                }
            }
        }
#endif
}

void MMRE_distributed(const double *ma,const double *mb, double *mc,int mn, int numproc, int cd, int blocksize)
{
    const double * a;
    const double * b;
    double * c;
    int n;
    int blk_size;
    int step;
    int substep;
    int maxij;
    int maxk;
    int iastart;
    int sqrtproc;
    int hdim;
    int mu;
    double alpha;
    int i,j,bj,p,stepi,stepj,k,ia,bi,bbj;

    /*copying parameters in private variables*/ 
    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;

    //need to find parameters mu
    mu = get_mu(cd);

    sqrtproc = (int)floor(sqrt(numproc));

    /*mu must be a multiple of matrix order:*/
    mu = mn / ((int)ceil((double)mn/(double)(sqrtproc*blocksize*mu)) * sqrtproc * blocksize);
    mu = max(mu,1);

    int lmb = mu;
    while(lmb>0 && n%(lmb*blocksize*sqrtproc) !=0 )
    {
        lmb--;
    }


    mu = max(lmb,1);

    step = sqrtproc*mu;
    //compute maxi and maxj (upper integer part)
    maxij = (int)ceil((double)n/((double)step*(double)blocksize));
    maxk = (int)floor((double)n/((double)blocksize));
    hdim = mu*blocksize;

    alpha = 1.0;
#ifdef ROWMAJOR
#pragma omp parallel firstprivate(alpha,hdim,a,b,c,mu,sqrtproc,step,maxij,maxk,n,blk_size) private(p,stepi,stepj,i,j,bj,substep,ia,bi,k,bbj) 
    {
#pragma omp for schedule(static,1) 
        for(p = 1; p<=numproc ;++p)
        {
            for(stepi = 0; stepi <maxij; ++stepi)
            {
                i = stepi*step;
                bi = i + ((p-1) % sqrtproc)*mu;

                for(stepj = 0; stepj <maxij; ++stepj)
                {
                    j = stepj*step;
                    bj=j+((p-1) / sqrtproc)*mu;

                    for(k=0; k<maxk;++k)
                    {
                        for(substep=0; substep<mu;++substep)
                        {
                            ia = bi + substep;
                            for(bbj=0;bbj<mu;++bbj){
                                cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,alpha,&A(ia*blk_size,k*blk_size),n,&B(k*blk_size,(bj+bbj)*blk_size),n,alpha,&C(ia*blk_size,(bj+bbj)*blk_size),n);
#ifdef SYNC
#pragma omp barrier
#endif
                            }
                        }
                    }
                }
            }
        }
    }
#else
#pragma omp parallel firstprivate(alpha,hdim,a,b,c,mu,sqrtproc,step,maxij,maxk,n,blk_size) private(p,stepi,stepj,i,j,bj,substep,ia,bi,k,bbj) 
    {
#pragma omp for schedule(static,1) 
        for(p = 1; p<=numproc ;++p)
        {
            for(stepj = 0; stepj <maxij; ++stepj)
            {
                j = stepj*step;
                bj=j+((p-1) / sqrtproc)*mu;
                for(stepi = 0; stepi <maxij; ++stepi)
                {
                    i = stepi*step;
                    bi = i + ((p-1) % sqrtproc)*mu;


                    for(k=0; k<maxk;++k)
                    {
                        for(substep=0; substep<mu;++substep)
                        {
                            ia = bi + substep;
                            for(bbj=0;bbj<mu;++bbj){
                                cblas_dgemm(CblasColMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,alpha,&A(ia*blk_size,k*blk_size),n,&B(k*blk_size,(bj+bbj)*blk_size),n,alpha,&C(ia*blk_size,(bj+bbj)*blk_size),n);
                            }
                        }
                    }
                }
            }
        }
    }
#endif
}




void Toledo_shared(const double *ma,const double *mb, double *mc,int mn, int numproc, int cs, int blocksize)
{
    const double * a;
    const double * b;
    double * c;
    int n;
    int blk_size;
    int macroblocksize;
    int sqrtproc;
    int maxijk;
    double alpha;
    int step;
    int substep;
    int p,i,j,k,stepi,stepj,pk,li,bi,lj,bj,lk;

    /*copying parameters in private variables*/ 
    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;


    macroblocksize = max((int)floor(sqrt((double)cs/3.0)),1);

    sqrtproc = (int)floor(sqrt(numproc));

    /*macroblocksize must be a multiple of sqrtproc and matrix order:*/
    int lmb = macroblocksize;
    while(lmb>0 && (lmb % sqrtproc != 0 || n%(lmb*blocksize) !=0) )
    {
        lmb--;
    }


    macroblocksize = max(lmb,sqrtproc);

    step = macroblocksize;
    substep = step / sqrtproc;

    maxijk = (int)ceil((double)n/((double)step*(double)blocksize));


    alpha = 1.0;
#ifdef ROWMAJOR
#pragma omp parallel firstprivate(alpha,a,b,c,sqrtproc,n,maxijk,blk_size,substep,step) private(i,j,bj,bi,p,stepi,stepj,k,pk,lk,li,lj) 
    {
#pragma omp for schedule(static,1) 
        for(p = 1; p<=numproc;++p){
            for(stepi = 0; stepi <maxijk; ++stepi)
            {
                i = stepi*step;

                for(stepj = 0; stepj <maxijk; ++stepj)
                {
                    //compute global j 
                    j = stepj*step;         

                    for(pk=0; pk<maxijk;++pk)
                    {
                        k=pk*step;

                        for(li=0;li<substep;++li)
                        {
                            bi=i+li + ((p-1) % sqrtproc )*substep;
                            for(lj=0;lj<substep;++lj)
                            {
                                bj=j+lj+((p-1) / sqrtproc )*substep;
                                for(lk=0;lk<step;++lk)
                                {
                                    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,alpha,&A(bi*blk_size,(k+lk)*blk_size),n,&B((k+lk)*blk_size,bj*blk_size),n,alpha,&C(bi*blk_size,bj*blk_size),n);
#ifdef SYNC
#pragma omp barrier
#endif
                                }
                            }
                        }
                    }
                }
            }
        }
    }
#else
#pragma omp parallel firstprivate(alpha,a,b,c,sqrtproc,n,maxijk,blk_size,substep,step) private(i,j,bj,bi,p,stepi,stepj,k,pk,lk,li,lj) 
    {
#pragma omp for schedule(static,1) 
        for(p = 1; p<=numproc;++p){ 
            for(stepj = 0; stepj <maxijk; ++stepj)
            {
                //compute global j 
                j = stepj*step;         

                for(stepi = 0; stepi <maxijk; ++stepi)
                {
                    i = stepi*step;
                    for(pk=0; pk<maxijk;++pk)
                    {
                        k=pk*step;

                        for(li=0;li<substep;++li)
                        {
                            bi=i+li + ((p-1) % sqrtproc )*substep;
                            for(lj=0;lj<substep;++lj)
                            {
                                bj=j+lj+((p-1) / sqrtproc )*substep;
                                for(lk=0;lk<step;++lk)
                                {
                                    cblas_dgemm(CblasColMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,alpha,&A(bi*blk_size,(k+lk)*blk_size),n,&B((k+lk)*blk_size,bj*blk_size),n,alpha,&C(bi*blk_size,bj*blk_size),n);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
#endif
}



void Toledo_distributed(const double *ma,const double *mb, double *mc,int mn, int numproc, int cd, int blocksize)
{
    const double * a;
    const double * b;
    double * c;
    int n;
    int blk_size;
    int macroblocksize;
    int sqrtproc;
    int maxij;
    int maxk;
    double alpha;
    int step;
    int substep;
    int p,i,j,stepk,stepi,stepj,my_i,my_j,li,lj,lk;
    /*copying parameters in private variables*/ 
    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;


    macroblocksize = max((int)floor(sqrt((double)cd/3.0)),1);

    sqrtproc = (int)floor(sqrt(numproc));


    /*sqrtproc*macroblocksize must be a multiple of matrix order:*/
    //  macroblocksize = mn / ((int)ceil((double)mn/(double)(blocksize* sqrtproc * macroblocksize )) * sqrtproc * blocksize);
    //  macroblocksize = max(macroblocksize,1);
    int lmb = macroblocksize;
    while(lmb>0 && n%(sqrtproc*lmb*blocksize) !=0 )
    {
        lmb--;
    }

    macroblocksize = max(lmb,1);


    step = sqrtproc*macroblocksize;

    maxij = (int)ceil((double)n/((double)step*(double)blocksize));
    maxk = (int)ceil((double)n/((double)macroblocksize*(double)blocksize));


    alpha = 1.0;
#ifdef ROWMAJOR
#pragma omp parallel firstprivate(alpha,a,b,c,sqrtproc,n,maxij,maxk,blk_size,step) private(my_i,my_j,i,j,p,stepi,stepj,stepk,li,lj,lk) 
    {
#pragma omp for schedule(static,1) 
        for(p = 1; p<=numproc;++p){
            my_i =  ((p-1) % sqrtproc )*macroblocksize;
            my_j = ((p-1) / sqrtproc )*macroblocksize;
            for(stepi = 0; stepi <maxij; ++stepi)
            {
                i = stepi*step;

                for(stepj = 0; stepj <maxij; ++stepj)
                {
                    //compute global j 
                    j = stepj*step;         

                    for(stepk=0; stepk<maxk;++stepk)
                    {
                        for(li=0; li<macroblocksize;++li)
                        {
                            for(lj=0; lj<macroblocksize;++lj)
                            {

                                for(lk=0; lk<macroblocksize;++lk)
                                {

                                    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,alpha,&A((i+my_i + li )*blk_size,(stepk*macroblocksize + lk)*blk_size),n,&B((stepk*macroblocksize + lk)*blk_size,(j+my_j + lj)*blk_size),n,alpha,&C((i+my_i + li)*blk_size,(j+my_j + lj)*blk_size),n);
#ifdef SYNC
#pragma omp barrier
#endif
                                }
                            }
                        }
                    }
                }
            }
        }
    }
#else
#pragma omp parallel firstprivate(alpha,a,b,c,sqrtproc,n,maxij,maxk,blk_size,step) private(my_i,my_j,i,j,p,stepi,stepj,stepk) 
    {
#pragma omp for schedule(static,1) 
        for(p = 1; p<=numproc;++p){
            my_i =  ((p-1) % sqrtproc );
            my_j = ((p-1) / sqrtproc );
            for(stepj = 0; stepj <maxij; ++stepj)
            {
                //compute global j 
                j = stepj*step;         
                for(stepi = 0; stepi <maxij; ++stepi)
                {
                    i = stepi*step;

                    for(stepk=0; stepk<maxk;++stepk)
                    {
                        for(lj=0; lj<macroblocksize;++lj)
                        {
                            for(li=0; li<macroblocksize;++li)
                            {

                                for(lk=0; lk<macroblocksize;++lk)
                                {

                                    cblas_dgemm(CblasColMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,alpha,&A((i+my_i + li )*blk_size,(stepk*macroblocksize + lk)*blk_size),n,&B((stepk*macroblocksize + lk)*blk_size,(j+my_j + lj)*blk_size),n,alpha,&C((i+my_i + li)*blk_size,(j+my_j + lj)*blk_size),n);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

#endif
}

void MMRE_Tradeoff_Modulo(const double *ma,const double *mb, double *mc,int mn, int numproc, int cs, int cd, double coms, double comd, int blocksize)
{
    const double * a;
    const double * b;
    double * c;
    int n;
    int mu;
    int sqrtproc;
    int tmp;
    int blk_size;
    int alphamax,alphamin,alpha,beta;
    int alphap, maxstep,maxsubij;
    int step,maxij;
    int i,j,p,stepi,stepj,k,ia,substep;
    int my_i,my_j,subi,subj,ka,ja,hdim;

    int it,jt,ig,jg,im,jm,ib,jb;

    /*copying parameters in private variables*/ 
    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;


    //need to find parameters lambda and mu
    mu = get_mu(cd);

    sqrtproc = (int)floor(sqrt(numproc));
    tmp = mu;
    while(tmp>0 && n%(tmp*sqrtproc*blocksize) !=0 )
    {
        tmp--;
    }

    mu = max(tmp,1);

    alphamax = get_alpha_max(cs);
    alphamin = sqrtproc*mu;


    get_alpha_beta(cs, coms, comd, numproc, mu, blocksize,n,&alpha,&beta);

    alphap = alpha/sqrtproc;


    int alphablk = alpha*blk_size;
    int giga_size = alpha/(sqrtproc*mu);
    int mega_size = mu;


    maxsubij = max(alpha/alphamin,1.0);

    step = alpha;
    hdim = mu*blk_size;

    maxij = (int)ceil((double)n/((double)step*(double)blocksize));
    maxstep = (int)ceil((double)n/((double)beta*(double)blocksize));



#ifdef ROWMAJOR
    double * tmp_c = malloc(alpha*blk_size*alpha*blk_size*sizeof(double));
#pragma omp parallel for schedule(static,1) firstprivate(alphablk,hdim,maxij,maxstep,maxsubij,step,beta,alpha,alphap,mu,blk_size,n,sqrtproc,giga_size,mega_size) private(p,stepi,stepj,i,j,substep,k,subi,subj,ka,ia,ja,my_i,my_j,it,jt,ig,jg,im,jm,ib,jb)
        for(p = 1; p<=numproc ;++p)
        {
            my_i =  ((p-1) % sqrtproc);
            my_j =  ((p-1) / sqrtproc);
/************* tera block loops*/
            for(it = 0; it < maxij; ++it)
            {
                for(jt = 0; jt <maxij; ++jt)
                {
                    for(ig = 0; ig < giga_size; ++ig)
                    {
                        for(jg = 0; jg <giga_size; ++jg)
                        {
                            for(im = 0; im < mega_size; ++im)
                            {
                                for(jm = 0; jm <mega_size; ++jm)
                                {
                                    for(ib = 0; ib < blk_size; ++ib)
                                    {
                                        for(jb = 0; jb <blk_size; ++jb)
                                        {
                                            int ilx = ib + blk_size*im + blk_size*mu*ig + my_i * alphap*blk_size + it * step;
                                            int jlx = jb + blk_size*jm + blk_size*mu*jg + my_j * alphap*blk_size + jt * step;
                                            int idxtmp = (my_i*sqrtproc + my_j)*alphap*blk_size*alphap*blk_size + (ig * giga_size + jg)*mu*blk_size*mu*blk_size + (im*mu + jm)*blk_size*blk_size + ib*blk_size + jb;
                                            tmp_c[idxtmp] = C(ilx,jlx);
#ifdef SYNC
#pragma omp barrier
#endif                                   
                                        }
                                    }
                                }
                            }
                        }
                    }
/*** computation loop ********/

                    for(substep=0; substep<maxstep;++substep){

                        k= substep*beta;

                        for(ig = 0; ig < giga_size; ++ig)
                        {
                            for(jg = 0; jg <giga_size; ++jg)
                            {
                                for(ka = k; ka <k + beta; ++ka)
                                {
                                    for(im = 0 ; im<mu; ++im)
                                    {
                                        
                                        ia = it *step + my_i*alphap + ig*mu + im;
                                        
                                        for(jm = 0 ; jm<mu; ++jm)
                                        {
                                        
                                            ja = jt *step + my_j*alphap + jg*mu + jm;
                                                                                double one = 1.0;

                                        int idxtmp = (my_i*sqrtproc + my_j)*alphap*blk_size*alphap*blk_size + (ig * giga_size + jg)*mu*blk_size*mu*blk_size + (im*mu + jm)*blk_size*blk_size;
                                           cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,one,&A(ia*blk_size,ka*blk_size),n,&B(ka*blk_size,ja*blk_size),n,one,&tmp_c[idxtmp],blk_size);
#ifdef SYNC
#pragma omp barrier
#endif                                   
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

        }
        free(tmp_c);
#endif
}
void MMRE_Tradeoff(const double *ma,const double *mb, double *mc,int mn, int numproc, int cs, int cd, double coms, double comd, int blocksize)
{
    const double * a;
    const double * b;
    double * c;
    int n;
    int mu;
    int sqrtproc;
    int tmp;
    int blk_size;
    int alphamax,alphamin,alpha,beta;
    int alphap, maxstep,maxsubij;
    int step,maxij;
    int i,j,p,stepi,stepj,k,ia,substep;
    int my_i,my_j,subi,subj,ka,ja,hdim;

    int it,jt,ig,jg,im,jm,ib,jb;

    /*copying parameters in private variables*/ 
    a = ma;
    b = mb;
    c = mc;
    n = mn;
    blk_size = blocksize;


    //need to find parameters lambda and mu
    mu = get_mu(cd);

    sqrtproc = (int)floor(sqrt(numproc));
    tmp = mu;
    while(tmp>0 && n%(tmp*sqrtproc*blocksize) !=0 )
    {
        tmp--;
    }

    mu = max(tmp,1);

    alphamax = get_alpha_max(cs);
    alphamin = sqrtproc*mu;
    get_alpha_beta(cs, coms, comd, numproc, mu, blocksize,n,&alpha,&beta);

    alphap = alpha/sqrtproc;

    int alphablk = alpha*blk_size;
    int giga_size = alpha/(sqrtproc*mu);
    int mega_size = mu;


    maxsubij = max(alpha/alphamin,1.0);

    step = alpha;
    hdim = mu*blk_size;

    maxij = (int)ceil((double)n/((double)step*(double)blocksize));
    maxstep = (int)ceil((double)n/((double)beta*(double)blocksize));


#ifdef ROWMAJOR
#pragma omp parallel for schedule(static,1) firstprivate(alphablk,hdim,maxij,maxstep,maxsubij,step,beta,alpha,alphap,mu,blk_size,n,sqrtproc,giga_size,mega_size) private(p,stepi,stepj,i,j,substep,k,subi,subj,ka,ia,ja,my_i,my_j,it,jt,ig,jg,im,jm,ib,jb)
        for(p = 1; p<=numproc ;++p)
        {
            my_i =  ((p-1) % sqrtproc);
            my_j =  ((p-1) / sqrtproc);
         
            for(stepi = 0; stepi <maxij; ++stepi)
            {
                i = stepi*step;

                for(stepj = 0; stepj <maxij; ++stepj)
                {
                    j = stepj*step;
                    for(substep=0; substep<maxstep;++substep){

                        k= substep*beta;

                        for(subi = 0; subi <maxsubij; ++subi)
                        {
                            for(subj = 0; subj <maxsubij; ++subj)
                            {
                                for(ka = k; ka <k + beta; ++ka)
                                {
                                    for(ia = i + my_i*alphap + subi*mu ; ia<i + my_i*alphap + (subi+1)*mu; ++ia)
                                    {
double one = 1.0;
                                        for(ja = j + my_j*alphap + subj*mu ; ja<j + my_j*alphap + (subj+1)*mu; ++ja)
                                        {
                                            cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,blk_size,blk_size,blk_size,one,&A(ia*blk_size,ka*blk_size),n,&B(ka*blk_size,ja*blk_size),n,one,&C(ia*blk_size,ja*blk_size),n);
#ifdef SYNC
#pragma omp barrier
#endif                                            
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

        }

#else
#pragma omp parallel for schedule(static,1) firstprivate(hdim,maxij,maxstep,maxsubij,step,beta,alphap,mu,blk_size,n) private(p,stepi,stepj,i,j,substep,k,subi,subj,ka,ia,ja,my_i,my_j)
        for(p = 1; p<=numproc ;++p)
        {
            my_i =  ((p-1) % sqrtproc);
            my_j =  ((p-1) / sqrtproc);
            for(stepj = 0; stepj <maxij; ++stepj)
            {
                j = stepj*step;


                for(stepi = 0; stepi <maxij; ++stepi)
                {
                    i = stepi*step;

                    for(substep=0; substep<maxstep;++substep){

                        k= substep*beta;
                        for(subj = 0; subj <maxsubij; ++subj)
                        {

                            for(subi = 0; subi <maxsubij; ++subi)
                            {
                                for(ka = k; ka <k + beta; ++ka)
                                {
                                    for(ia = i + my_i*alphap + subi*mu ; ia<i + my_i*alphap + (subi+1)*mu; ++ia)
                                    {
                                        for(ja = j + my_j*alphap + subj*mu ; ja<j + my_j*alphap + (subj+1)*mu; ++ja)
                                        {
                                            dgemm("N","N",&blk_size,&blk_size,&blk_size,&alpha,&A(ia*blk_size,ka*blk_size),&n,&B(ka*blk_size,ja*blk_size),&n,&alpha,&C(ia*blk_size,ja*blk_size),&n);
                                        }
#ifdef SYNC
#pragma omp barrier
#endif                                            
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }        
#endif
}


int main (int argc, char * const argv[]) {
    int i, info, j;
    double* a;
    double* b;
    double* c;
    double* d;
    double sum;
    struct timeval tvbeg;
    struct timeval tvend;
    int blocksize = 96;
    enum algo_type algorithm_type= OUTER;
    int n = 0;
    int numproc;
    int numcore;
    double elapsed;
    struct bitmask * cpumask;
    struct multicore_topo topology;
    int err;
    int cs = 0;
    int cd = 0;

    if(numa_available()==-1) exit(-1);

    read_arguments(argc,argv,&blocksize,&algorithm_type,&n);

    err = ReadTopology(&topology);
    if(err){
        exit(-1);
    }

    numproc = numa_max_node()+1;
numproc = topology.node_count;
 
    numcore = omp_get_num_procs();
numcore = topology.core_count / topology.node_count;
    omp_set_num_threads(numcore);

    /*convert cs from kB to number of double*/
    cs = topology.cs * 1024 / sizeof(double);
    /*convert cs from doubles to number of block of blocksize * blocksize doubles*/
    cs /= blocksize*blocksize;
    cs *= CACHE_PROP_CS;
    /*convert cd from kB to number of double*/
    cd = topology.cd * 1024 / sizeof(double);
    /*convert cd from doubles to number of block of blocksize * blocksize doubles*/
    cd /= blocksize*blocksize;
    cd *= CACHE_PROP_CD;

    printf("%d processors of %d cores cs %d cd %d\n",numproc,numcore,cs,cd);

    posix_memalign(&a,64,sizeof(double)*n*n);
    posix_memalign(&b,64,sizeof(double)*n*n);
    posix_memalign(&c,64,sizeof(double)*n*n);

#ifdef CHECK
    posix_memalign(&d,64,sizeof(double)*n*n);
#endif
    for (i = 0; i < n; i++)
    {
        for (j = 0; j < n; j++)
        {
            A(i,j) = i*2+j;
            B(i,j) = j*3+i;
        }
    }
#ifdef CHECK
    memset(c,0,n*n*sizeof(double));

    {
        double alpha = 1.0;
#ifdef PARA_BLAS
        printf("PARALLEL LIBRARY BLAS BEGIN\n");
#else
        printf("LIBRARY BLAS BEGIN\n");
#endif
        gettimeofday(&tvbeg,NULL);
#ifdef ROWMAJOR
        cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,n,n,n,alpha,&A(0,0),n,&B(0,0),n,alpha,&C(0,0),n);
#else
        cblas_dgemm(CblasColMajor,CblasNoTrans,CblasNoTrans,n,n,n,alpha,&A(0,0),n,&B(0,0),n,alpha,&C(0,0),n);
#endif
        gettimeofday(&tvend,NULL);

#ifdef PARA_BLAS
        printf("PARALLEL LIBRARY BLAS END\n");
#else
        printf("LIBRARY BLAS END\n");
#endif
    }

    memcpy(d,c,n*n*sizeof(double));
#endif
    memset(c,0,n*n*sizeof(double));


    switch (algorithm_type) {
        case LIBRARY:
            {
                double alpha = 1.0;
#ifdef PARA_BLAS
                printf("PARALLEL LIBRARY BLAS BEGIN\n");
#else
                printf("LIBRARY BLAS BEGIN\n");
#endif
                gettimeofday(&tvbeg,NULL);
#ifdef ROWMAJOR
                cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,n,n,n,alpha,&A(0,0),n,&B(0,0),n,alpha,&C(0,0),n);
#else
                cblas_dgemm(CblasColMajor,CblasNoTrans,CblasNoTrans,n,n,n,alpha,&A(0,0),n,&B(0,0),n,alpha,&C(0,0),n);
#endif
                gettimeofday(&tvend,NULL);

#ifdef PARA_BLAS
                printf("PARALLEL LIBRARY BLAS END\n");
#else
                printf("LIBRARY BLAS END\n");
#endif
            }
            break;
        case OUTER:
            {
                printf("OUTER PRODUCT BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                matrix_multiplication(a,b,c,n, numcore,blocksize);
                gettimeofday(&tvend,NULL);
                printf("OUTER PRODUCT END\n");
            }
            break;
        case TOLEDO:
            {
                printf("TOLEDO DIST BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                Toledo_distributed(a,b,c,n, numcore, CACHE_PROP_CD*(double)cd, blocksize);
                gettimeofday(&tvend,NULL);
                printf("TOLEDO DIST END\n");
            }
            break;
        case TOLEDO_LRU:
            {
                printf("TOLEDO DIST LRU BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                Toledo_distributed(a,b,c,n, numcore,(int) (0.5 *CACHE_PROP_CD*(double) cd), blocksize);
                gettimeofday(&tvend,NULL);
                printf("TOLEDO DIST LRU END\n");
            }
            break;

        case TOLEDO_S:
            {
                printf("TOLEDO SHARED BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                Toledo_shared(a,b,c,n, numcore, CACHE_PROP_CS*(double)cs, blocksize);
                gettimeofday(&tvend,NULL);
                printf("TOLEDO SHARED END\n");
            }
            break;
        case TOLEDO_SLRU:
            {
                printf("TOLEDO SHARED LRU BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                Toledo_shared(a,b,c,n, numcore,(int)(0.5*CACHE_PROP_CS*(double)cs), blocksize);
                gettimeofday(&tvend,NULL);
                printf("TOLEDO SHARED LRU END\n");
            }
            break;
        case MMRE_S:
            {
                printf("MMRE SHARED BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_shared(a,b,c,n, numcore, CACHE_PROP_CS*(double)cs, blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE SHARED END\n");
            }
            break;


        case MMRE_SLRU:
            {
                printf("MMRE SHARED LRU BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_shared(a,b,c,n, numcore, (int)(0.5*CACHE_PROP_CS*(double)cs), blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE SHARED LRU END\n");
            }
            break;

        case MMRE_D:
            {
                printf("MMRE DIST BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_distributed(a,b,c,n, numcore, CACHE_PROP_CD*(double)cd, blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE DIST END\n");
            }
            break;

        case MMRE_DLRU:
            {
                printf("MMRE DIST LRU BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_distributed(a,b,c,n, numcore, (int)(0.5*CACHE_PROP_CD*(double)cd), blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE DIST LRU END\n");
            }
            break;

        case MMRE_TM:
            {
                printf("MMRE TRADEOFF MODULO BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_Tradeoff_Modulo(a,b,c,n,numcore,CACHE_PROP_CS*(double)cs,CACHE_PROP_CD*(double)cd, 2429340876 , 4563402752,blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE TRADEOFF MODULO END\n");
 

            }
            break;

        case MMRE_TMLRU:
            {
                 printf("MMRE TRADEOFF MODULO LRU BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_Tradeoff_Modulo(a,b,c,n,numcore,0.5*CACHE_PROP_CS*(double)cs,0.5*CACHE_PROP_CD*(double)cd, 2429340876 , 4563402752,blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE TRADEOFF MODULO LRU END\n");
 


            }
            break;

        case MMRE_T:
            {
                printf("MMRE TRADEOFF BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_Tradeoff(a,b,c,n,numcore,CACHE_PROP_CS*(double)cs,CACHE_PROP_CD*(double)cd, 2429340876 , 4563402752,blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE TRADEOFF END\n");
 

            }
            break;

        case MMRE_TLRU:
            {
                 printf("MMRE TRADEOFF LRU BEGIN\n");
                gettimeofday(&tvbeg,NULL);
                MMRE_Tradeoff(a,b,c,n,numcore,0.5*CACHE_PROP_CS*(double)cs,0.5*CACHE_PROP_CD*(double)cd, 2429340876 , 4563402752,blocksize);
                gettimeofday(&tvend,NULL);
                printf("MMRE TRADEOFF LRU END\n");
 


            }
            break;
        default:
            break;
    }

#ifdef CHECK
    for(i=0; i<n;i++){
        for(j=0;j<n;j++){
            if(C(i,j)!=D(i,j)){
                printf("ERREUR CALCUL C(%d,%d)=%lf  D(%d,%d)=%lf \n\n",i,j,C(i,j),i,j,D(i,j));
                return -1;
            }
        }
    }
#endif

    elapsed = (double)tvend.tv_sec * 1000.0 + (double)tvend.tv_usec / 1000.0;
    elapsed -= (double)tvbeg.tv_sec * 1000.0 + (double)tvbeg.tv_usec / 1000.0;
    printf("Elapsed time %lf.3 ms\n",elapsed);

#ifdef CHECK
    free(d);
#endif
    free(c);
    free(b);
    free(a);

    return 0;
}

